# == Schema Information
# Schema version: 20090930163041
#
# Table name: contents
#
#  id              :integer(4)      not null, primary key
#  otu_id          :integer(4)
#  content_type_id :integer(4)
#  text            :text
#  is_public       :boolean(1)      default(TRUE), not null
#  pub_content_id  :integer(4)
#  revision        :integer(4)
#  proj_id         :integer(4)      not null
#  creator_id      :integer(4)      not null
#  updator_id      :integer(4)      not null
#  updated_on      :timestamp       not null
#  created_on      :timestamp       not null
#

class Content < ActiveRecord::Base
  has_standard_fields

  include ModelExtensions::Taggable
  include ModelExtensions::Figurable
  include ModelExtensions::DefaultNamedScopes
  include ModelExtensions::MiscMethods

  belongs_to :content_type
  belongs_to :otu

  has_one :public_version, :class_name => 'PublicContent', :foreign_key => 'pub_content_id', :dependent => :destroy

  named_scope :private, :conditions => "pub_content_id is null" 
  named_scope :by_otu, lambda {|*args| {:conditions => ["contents.otu_id = ?", args.first || -1] }}
  named_scope :by_content_type, lambda {|*args| {:conditions => ["contents.content_type_id = ?", args.first || -1] }}

  validates_presence_of :text

  # doesn't require :otu and :content_type because of public_content, should perhaps be subclassed

  def validate
    errors.add(:otu, 'content must be attached to a OTU') if !otu && pub_content_id.blank?
    errors.add(:content_type_id, 'content must be have a content type') if !content_type && pub_content_id.blank?
  end

  def display_name(options = {})
     self.text.slice(0..300) + (self.text.size > 300 ? "..." : '')
  end
  
  def is_published
    self.public_version  ? true : false
  end

  def publish

    begin
      Content.transaction do

        if p = self.public_version
          # check to see if text is the same- if it is don't increment the revision
          if not self.text == p.text
            p.update_attributes(:revision => p.revision + 1, :text => self.text)    
          end
        
          # wipe the public figures
          for f in p.figures
            f.destroy
          end
          
        else
          p = PublicContent.new
          p.text = self.text
          p.pub_content_id = self.id
          p.revision = 1
          p.save!
        end
       
        # recreate the figures
        for f in self.figures
          fnew = f.clone
          fnew.addressable_id = p.id
          fnew.save!
        end
         
      end

      rescue Exception => e
        flash[:notice] = e.message 
        return false
      end
    return true
  end

  # transfer this content to a different Otu, if delete_from_incoming = false then the content (and figures, tags) is duplicated
  def transfer_to_otu(otu, delete_from_incoming = true)
    return false if otu.id == self.otu_id
    c = Content.find(:first, :conditions => {:otu_id => otu.id, :content_type_id => self.content_type_id})
    # does the incoming otu have this content_type already
    
    begin
      Content.transaction do
        if c # content already exists for the clone-to OTU 
          c.text = c.text + " [Transfered from OTU #{self.otu_id}: #{self.text}]"  # append content
          c.save! 
          # deal with figures
          for f in self.figures
            if found = Figure.find(:first, :conditions => {:addressable_type => 'Content', :addressable_id => c.id , :image_id => f.image_id})
              # append the caption
              found.caption +=  " [Transfered from OTU #{self.otu_id}: #{f.caption}]"
              f.destroy if delete_from_incoming
            else
              if delete_from_incoming
                # transfer the figure
                f.addressable_id = c.id
                f.save!
              else
                # clone the figure
                n = f.clone
                n.addressable_id = c.id
              end
            end
          end

          # deal with tags
          for t in self.tags
            if found = Tag.find(:first, :conditions => {:addressable_type => 'Content' , :addressable_id => c.id , :keyword_id => t.keyword_id })
              # append the notes
              
              found.notes += " [Transfered from OTU #{self.otu_id}: #{t.notes}]"
              found.save!
              t.destroy if delete_from_incoming
            else
              if delete_from_incoming
                # transfer the tag
                t.addressable_id = c.id
                t.save!
              else
                #clone the tag
                n = t.clone
                n.addressable_id = c.id
                n.save!
              end
            end
          end

          self.destroy if delete_from_incoming
   
        else  #  content in transfer-to OTU doesn't exist
          if delete_from_incoming  # just update the otu_id and everything gets transferred
            self.otu_id = otu.id
            self.save!
          else # need to clone everything
            # clone the content
            nc = self.clone
            nc.otu_id = otu.id
            nc.save!

            # clone figures
            for f in self.figures
              fn = f.clone
              fn.addressable_id = nc.id
              fn.save!
             end

            # clone tags
            for t in self.tags
              tn = t.clone
              tn.addressable_id = nc.id
              tn.save!
            end
          end
        end
      end

    rescue
      returns false
    end
    true
  end


end
