class Label < ActiveRecord::Base
  # Labels are names for OntologyClasses, their instantiation implies their association with at least one instance of a "real life" OntologyClass whether or not that association has be captured in the database
  has_standard_fields
  include ModelExtensions::Taggable
  include ModelExtensions::DefaultNamedScopes

  belongs_to :language
  belongs_to :plural_of_label, :foreign_key => 'plural_of_label_id', :class_name => "Label"

  has_one :plural_form, :foreign_key => 'plural_of_label_id', :class_name => "Label"

  has_many :sensus, :dependent => :destroy
  has_many :ontology_classes, :through => :sensus, :uniq => true # directly tied, plural ties are not included here
  has_many :ontology_classes_as_plural, :through => :sensus, :primary_key => 'plural_of_label_id' # #TEST directly tied, plural ties are not included here
  has_many :ontology_classes_in_OBO, :foreign_key => 'obo_label_id', :class_name => 'OntologyClass'

  named_scope :with_definitions, :conditions => "labels.id IN (SELECT label_id FROM sensus)"  # alias for with_ontology_classes
  named_scope :without_ontology_classes, :conditions => "labels.id NOT IN (SELECT label_id FROM sensus)" 
  named_scope :all_for_ontology_class, lambda {|*args| {:conditions => ["(labels.id in (select s.label_id from sensus s where s.ontology_class_id = ?)) OR (labels.id in (select lp.id from labels lp where lp.plural_of_label_id in (select s.label_id from sensus s where s.ontology_class_id = ?))) ", args.first || -1, args.first || -1]}}
  named_scope :ordered_by_label_length, :order => 'length(labels.name)'

  named_scope :singular, :conditions => 'plural_of_label_id is null'

  # pass an Array of strings, escape results *before* you use with_label_from_array 
  named_scope :with_label_from_array, lambda {|*args| {:conditions => (args.first.size > 0 ? "(" + args.first.collect{|a| "(labels.name = \"#{a.gsub(/\"/, "\"")}\")"}.join(" OR ") + ")" : "labels.name = '-1'") }}
  named_scope :with_more_than_one_ontology_class, :conditions => 'id IN (select id from sensus group by ontology_class_id, ref_id having count(id) > 1)'
  named_scope :with_first_letter, lambda {|*args| { :conditions => ["name LIKE ?", (args.first ? "#{args.first}%" : -1)]}} 
  named_scope :without_plural_forms, :conditions => 'id NOT IN (SELECT plural_of_label_id id FROM labels where plural_of_label_id IS NOT NULL)'

  validates_presence_of :name
  validates_format_of :name, :with => /\A\w+(\Z|.*\w+\Z)/i, :message => 'invalid label, you may need to trim space'
  validates_uniqueness_of :name, :scope => :proj_id, :message => 'that label is already present in this project'
  
  # UGH! why can't this be a  validates_uniqueness_of :plural_of_label_id, :scope => :id, :allow_blank => true #  [:id, :proj_id] ???
  before_validation :validate_plurals_are_not_further_pluralized
  # before_validation :validate_plurals_are_not_used_in_sensu
  before_destroy :validate_label_not_used_with_xrefed_ontology_class
  before_destroy :validate_label_not_used_as_obo_label
    
  def display_name(options = {})
     opt = {
      :type => nil # :list, :head, :select, :selected, :sub_select, :material_examined_verbose
     }.merge!(options.symbolize_keys)
     case opt[:type]
      when[:select]
        name
      else
        name
      end
  end

  def has_plural?
    !self.plural_form.blank?
  end

  def is_plural?
    !self.plural_of_label.blank?
  end

  # aliased
  def has_ontology_class?
    has_defintion?
  end

  def has_ontology_class_with_dbxref?
    self.ontology_classes.with_populated_dbxref.size > 0 
  end

  def has_definition?
    self.sensus.size > 0
  end

  # returns an Array of Strings
  def all_forms
    plural_form.blank? ? [name] : [plural_form.name, name].sort
  end

  # returns a Hash of {Label => [OntologyClass1...OntologyClassn]} where 
  # OntologyClass has Label.name in definition AND Label.ontology_classes.size == 0
  def self.without_ontology_classes_but_used_in_ontology_class_definitions(params)
    opts = {:proj_id => nil}.merge!(params) 
    if proj = Proj.find(opts[:proj_id])
      labels = {} 
      proj.labels.without_ontology_classes.each do |l|
        ocs = proj.ontology_classes.with_definition_containing(l.name)
        labels.merge!(l => ocs) if ocs.size > 0
      end
    else 
      return {} 
    end
    return labels 
  end 

  def synonyms_by_ontology_class(ontology_class)
    Sensu.by_ontology_class(ontology_class).excluding_label(self).collect{|s| s.label}.uniq.sort{|a,b| a.name <=> b.name}
  end

  def self.auto_complete_search_result(params = {})
     tag_id_str = params[:tag_id]
     return false if (tag_id_str == nil  || params[:proj_id].blank?)

     value = params[tag_id_str.to_sym].split.join('%')

     lim = case params[tag_id_str.to_sym].length
       when 1..2 then 3 
       when 3..4 then 5 
       else lim = false # no limits
     end 
     Label.find(:all, :conditions => ["(name LIKE ? OR id = ?) AND proj_id = ?", "%#{value}%", value.gsub(/\%/, ""), params[:proj_id]], :order => "length(name)", :limit => lim ).uniq
   end

   def self.search_redirect(params = {})
     {:controller => params[:hidden_field_class_name].downcase.to_sym, :id => params[:onto_search][:search_id], :action => :show}
   end

  protected

  # Validation
  def validate_plurals_are_not_further_pluralized 
    if !plural_of_label_id.blank? && Label.find_by_plural_of_label_id(plural_of_label_id)
      errors.add(:plural_of_label_id, "The root of this label is already pluralized.")
      false
    end
  end

  def validate_label_not_used_with_xrefed_ontology_class
    if self.ontology_classes.with_populated_dbxref.size > 0 
      errors.add(:name, "This label is used for a dbxrefed class or as an OBO label. It can not be destroyed.")
      false
    end
  end

  def validate_label_not_used_as_obo_label
    if self.ontology_classes_in_OBO.size > 0
      errors.add(:name, "This label is used as an OBO label. It can not be destroyed.")
      false
    end
  end

end
