# == Schema Information
# Schema version: 20090930163041
#
# Table name: ontologies
#
#  id         :integer(4)      not null, primary key
#  part1_id   :integer(4)      not null
#  part2_id   :integer(4)      not null
#  isa_id     :integer(4)      not null
#  notes      :text
#  proj_id    :integer(4)      not null
#  creator_id :integer(4)      not null
#  updator_id :integer(4)      not null
#  updated_on :timestamp       not null
#  created_on :timestamp       not null
#

# this Model tracks the relationships b/w two parts

class Ontology < ActiveRecord::Base
  require 'obo_parser'

  has_standard_fields
  belongs_to :part1, :foreign_key => 'part1_id', :class_name => 'Part'
  belongs_to :part2, :foreign_key => 'part2_id', :class_name => 'Part'
  
  belongs_to :isa
  
  validates_presence_of :part1
  validates_presence_of :part2 
  validates_presence_of :isa
 
  after_save :update_part_modified_on
  before_destroy :update_part_modified_on
 
  named_scope :by_relationship, lambda {|*args| {:conditions => ["isa_id = ?",  (args.first || -1)]}}

  #TODO: test
  named_scope :isa_relationships, :conditions => "isa_id in (SELECT id from isas where isas.name 'is_a')"
  named_scope :non_isa_relationships, :conditions => "isa_id in (SELECT id from isas where isas.name is not 'is_a')"

  # pass a Person TODO: roles, or extend to all ActiveRecord
  def created_or_admin(person)
    if person.id.to_i == self.creator_id.to_i || person.is_admin
      true
    else
      false
    end
  end

  # Maps an OBO file to a full ontology in mx, instantiating new Parts, Tags, and Ontologies where necessary
  # Returns a array of arrays, each of which contains a hash.  
  # Array index maps 1:1 with order of terms in incoming :file, it looks like this:
  # [ [ :part => Part, :tags => [ Tag1, ...Tagn], :ontologies => [Ontology1...Ontologyn], ... [] ]
  # Could class this out, but likely a waste of time.
  def self.terms_from_obo(options = {})
    @opt = {
      :file => nil,
      :proj_id => nil
    }.merge!(options.symbolize_keys)
    
    return false if (@opt[:file].blank? || @opt[:file].size == 0 )
    @proj = Proj.find(@opt[:proj_id])
    of = parse_obo_file(@opt[:file])
    result = {}
    obo_map  = {}

    of.terms.each_with_index do |term, i|

      obo_map[i] = {:part => nil, :tags => [], :ontologies => []}

      # the :part
      if @t = Part.find(:first, :conditions => {:name => term.name, :proj_id => @opt[:proj_id]}) 
        obo_map[i] = {:part => @t}
      else 
        obo_map[i] = {:part =>  Part.new(:name => term.name, 
                                        :description => (term.tags["def"].blank? ? nil : term.tags["def"].first))}
      end

      # the :tags
      # Tags need to have their addressable_id set post Part#save
      tags = []
      tags += [Tag.new(:addressable_type => 'Part', :keyword => @xref_kwd, :referenced_object => term.id)]

      # additional xrefs (Tags in mx)
      if !term.tags["xref"].blank?
        tags += term.tags["xref"].collect{|x| Tag.new(:addressable_type => 'Part', :keyword => @xref_kwd, :referenced_object => x) }
      end 
      obo_map[i].merge!(:tags => tags)
    end 
    
    # properties, only handles is a, part of now
    @isa = @proj.isas.by_interaction("is_a").first
    @part_of = @proj.isas.by_interaction("part_of").first
    if !@isa.blank? && !@part_of.blank?
    
      obo_map.keys.each do |i|
        obo_map[i].merge!(:ontologies => [])
        ["is_a", "part_of"].each do |rel|
          if !of.terms[i].tags[rel].blank?
            os = []  
             of.terms[i].tags[rel].each do |id|
               # part1 is self
               # part2 is index
                os << {:ontology =>  Ontology.new(:part1_id => nil, :part2_id => nil,  :isa_id => (rel == "is_a" ? @isa.id : @part_of.id)),
                       :part2_index => self._index_of_term_from_obo_map_for_term(of, id) }
             end
            obo_map[i][:ontologies] += os 
          end
        end
      end
    end

      # <OboFile::OboFile::Term:0x4d7daa8 @name="cell", @id="CL:0000000", @tags={"xref"=>["FMA:68646"], "def"=>["\"Anatomical structure that has as its parts a maximally connected cell compartment surrounded by a plasma membrane.\" [CARO:mah]"]}>

    result = obo_map
  end

  def self._index_of_term_from_obo_map_for_term(parsed_obo_file, xref)
    parsed_obo_file.terms.each_with_index do |t,i|
      return i if xref == t.id 
    end
    false
  end

  # these are default included in OBO
  def self.OBO_TYPEDEFS
   ['is_a', 'disjoint_from', 'instance_of', 'inverse_of', 'union_of', 'intersection_of']
  end
  
  # update the modified_on for the related parts, so we can keep track of what's being modified 
  def update_part_modified_on
    Part.find(part1_id).save!
    Part.find(part2_id).save!
  end
 
  def validate_on_create
   if Ontology.find_by_part1_id_and_part2_id_and_isa_id(part1_id, part2_id, isa_id)
       errors.add("That relationship/part combination already exists, it therefor ")
    end
  end
  
  def display_name(options = {})
    part1.display_name + " " + isa.display_name + " " + part2.display_name
  end

  def display_OBO_relationship_tag
    if self.isa.interaction == "is_a"
      "#{self.isa.interaction}: #{self.part2.obo_dbxref} ! #{self.part2.display_name}" 
    else
      "relationship: #{self.isa.interaction} #{self.part2.obo_dbxref} ! #{self.part2.display_name}" 
    end
  end

  def self.dot(options = {})
    g = GraphViz.new( :G, :type => :digraph )

   Proj.find(options[:proj_id]).ontologies(:include => [:part1, :part2, :isa]).each do |o|
      one = g.add_node( o.part1_id.to_s, "label" => o.part1.name )
      two = g.add_node( o.part2_id.to_s, "label" => o.part2.name )

      # Create an edge between the two nodes
      g.add_edge( one, two, "label" => o.isa.interaction, "color" => (o.isa.interaction == "is_a" ? "red" : "blue")  )
   end
    g   
  end

  protected

  def validate_destruction
  end 
end
