require 'ontology/visualize/newick'
require 'ontology/visualize/svg'

class OntologyClass < ActiveRecord::Base
  set_table_name "ontology_classes"
  has_standard_fields
  include ModelExtensions::Taggable
  include ModelExtensions::Figurable
  include ModelExtensions::DefaultNamedScopes
  include ModelExtensions::MiscMethods

  serialize :illustration_IP_votes, Array

  belongs_to :written_by, :foreign_key => 'written_by_ref_id', :class_name => 'Ref'
  belongs_to :obo_label,  :foreign_key => 'obo_label_id', :class_name => 'Label'
  belongs_to :taxon_name, :foreign_key => 'highest_applicable_taxon_name_id', :class_name => 'TaxonName'

  # primary_relationship object_relationship secondary_relationship, e.g. antennae (primary) part_of head (secondary)
  has_many :primary_relationships, :class_name => 'OntologyRelationship', :foreign_key => 'ontology_class1_id', :dependent => :destroy 
  has_many :secondary_relationships, :class_name => 'OntologyRelationship', :foreign_key => 'ontology_class2_id', :dependent => :destroy

  has_many :sensus, :dependent => :destroy
  has_many :labels, :through => :sensus, :uniq => true # singular only now, these are homonyms 
  has_many :all_labels_through_sensus, :through => :sensus, :source => :label

  # REVISIT: 
  # do is_obsolete classes in OBO have is_a relationships?

  # pass a String or Array of Strings, filters by singular or plural forms of Labels matching arguments
  named_scope :by_label_including_plurals, lambda {|*args| {
    :conditions => "ontology_classes.id IN (select id from 
    ((SELECT ontology_class_id id from sensus s  JOIN labels l ON s.label_id = l.id WHERE #{SqlHelper::where_scope_for_name(:l, *args)} )
    UNION (SELECT ontology_class_id id from sensus s2 JOIN labels l1 ON s2.label_id = l1.id JOIN labels lj on l1.id = lj.plural_of_label_id WHERE #{SqlHelper::where_scope_for_name(:lj, *args)})) alias_b)"} 
  }

  named_scope :with_definition_containing, lambda {|*args| {:conditions => ["definition REGEXP ?",  (args.first ?  "[[:<:]]#{args.first}[[:>:]]" : -1)]}} 
  named_scope :with_populated_dbxref, :conditions => "dbxref IS NOT NULL AND dbxref != ''"
  named_scope :with_dbxref_namespace, lambda {|*args| {:conditions => ["dbxref LIKE ?", args.first + ":%"]}} # :order => 'dbxref', 
  named_scope :that_are_obsolete, :conditions => "is_obsolete = 1"

  # if you pass a OntologyClass.id the scope is limited to those above it
  named_scope :without_ontology_relationships, lambda {|*args| {:conditions => "id NOT IN (SELECT ontology_class1_id from ontology_relationships) AND id NOT IN (SELECT ontology_class2_id from ontology_relationships)" }} 
  named_scope :without_sensus, lambda {|*args| {:conditions => "id NOT IN (SELECT ontology_class_id from sensus)"}} 
  named_scope :with_obo_label, :conditions => 'obo_label_id IS NOT NULL and obo_label_id != ""'
  named_scope :ordered_by_dbxref, :order => 'dbxref'

  validates_presence_of :definition
  validates_presence_of :written_by 
  validates_uniqueness_of :definition, :scope => :proj_id, :allow_blank => false, :allow_nil => false, :message => "That defintion already exists in this ontology."
  validates_uniqueness_of :dbxref, :message => 'xref alredy exists, pick a new one', :allow_blank => true, :allow_nil => true, :scope => "proj_id" 
  validates_format_of :dbxref, :with => /\A\w+\:\w+\Z/i, :message => 'must be in the format "foo:bar"', :if => Proc.new{|o| !o.dbxref.blank?} 
  validates_format_of :definition, :with => /\A\S/i, :message => 'invalid format for definition' # require there is a word character at the start, followed by something (TODO: make it end in non-whitespace)

  # TODO: is this safe? does written_by really equate to sensu?! (probably not)
  after_save :ensure_that_labels_contains_obo_label

  def illustration_IP_vote=(value)
    # validate and/or Raise
    self.illustration_IP_votes = [] if self.illustration_IP_votes == nil
    self.illustration_IP_votes << value
  end

  def before_save
    self.illustration_IP_votes = [] if self.illustration_IP_votes == nil
  end

  def before_destroy
    if !self.dbxref.nil?
      errors.add(:dbxref, "You can't destroy a class that has an existing dbxref in this manner.")
      false
    end
  end

  # TODO: Roles deprecation
  def created_or_admin(person)
    if person.id.to_i == self.creator_id.to_i || person.is_admin || person.is_ontology_admin
      true
    else
      false
    end
  end

  def display_name(options = {})
    opt = {:type => :inline 
    }.merge!(options.symbolize_keys)
    case opt[:type]
    when :select  
      "#{self.definition} <span style='padding:1px;background-color:#b5ebc7;'>#{self.label_name(:type => :preferred)}</span>"
    when :label_first
      "<span style='padding:1px;background-color:#b5ebc7;'>#{self.label_name(:type => :preferred)}</span> #{self.definition}"
    when :figure
      "#{self.definition} <span style='padding:1px;background-color:#b5ebc7;'>#{self.label_name(:type => :preferred)}</span>"
    when :inline
      definition 
    else
      definition 
    end  
  end

  def label_name(options = {})
    opt = {
      :type => :preferred 
    }.merge!(options.symbolize_keys)
    if self.labels.size > 0 || !self.obo_label
      case opt[:type]
      when :top_sensu
        self.sensus.ordered_by_position.first.label.name
      when :oldest
        self.sensus.ordered_by_age.first.label.name
      when :preferred
        if self.sensus.size > 0
          self.sensus.ordered_by_position.first.label.name           
        elsif self.obo_label
          self.obo_label.name
        end
      end
    else
        'no labels'
    end
  end

  # Array of Strings
  # includes plurals etc. 
  # TODO: refactor to a single SQL in Label?
  def all_labels
    self.labels.inject([]){|sum, l| sum + l.all_forms}.sort
  end

  # keys are Labels, values are an Array of ids, for mapping in Linker
  def hash_labels
    all_labels.inject({}){|h, l| h.merge(l => [self.id])}
  end

  def preferred_label
    self.sensus.ordered_by_position.first.label
  end

  def relationships
    OntologyRelationship.find(:all, :conditions => "proj_id = #{self.proj_id} AND ((ontology_class1_id = #{self.id}) OR (ontology_class2_id = #{self.id}))" )
  end

  # returns an Array of Strings
  # items are are 'Foo:1234' style references derived from Tags that use is_dbxref Keywords
  def dbxrefs_from_tags
    xrefs = [] 
    xrefs += self.tags.with_dbxref_keywords.collect{|x| !x.referenced_object.blank? ? x.referenced_object : nil}
    xrefs.compact.uniq
  end

  # returns Array of OntologyRelationships
  # immediately attached children only (does not recurse) 
  def tree_child(options = {})
    opt = { # recursing with @opt is bad
      :relationship_type => 'all'  # or an RelationshipObject#id
    }.merge!(options.symbolize_keys)

    # TODO: modifity to sort by first(top) label
    if opt[:relationship_type] == 'all'    
      OntologyRelationship.find(:all, :include => [:ontology_class1, :object_relationship, :ontology_class2], :conditions => ['ontology_class2_id = ?', self.id]).sort{|x,y| x.ontology_class1.definition <=> y.ontology_class1.definition}
    else
      OntologyRelationship.find(:all, :include => [:ontology_class1, :object_relationship, :ontology_class2], :conditions => ['ontology_class2_id = ? AND object_relationship_id = ?', self.id, opt[:relationship_type]]).sort{|x,y| x.ontology_class1.definition <=> y.ontology_class1.definition}
    end 
  end

  # A precursor example of finding all transitive relationships, in this case defaulted for is_a/part_of
  # returns a hash of OntologyClasses, with values == true when redundant relationships are implied 
  def logical_relatives(options = {})
    return [] if !ObjectRelationship.find_by_interaction_and_proj_id('is_a', self.proj_id) || !ObjectRelationship.find_by_interaction_and_proj_id('part_of', self.proj_id)

    rel1 = Proj.find(self.proj_id).object_relationships.by_interaction('is_a').first.id
    rel2 = Proj.find(self.proj_id).object_relationships.by_interaction('part_of').first.id
    return [] if !rel1 || !rel2 

    opt = { # recursing with @opt is bad
      :direction => :children,  # [:parents | :children]
      :rel1 => rel1,
      :rel2 => rel2 
      }.merge!(options.symbolize_keys)

      return nil if ![:parents, :children].include?(opt[:direction]) 

      first_result = []  # all the part_of
      second_result = [] # part_of through is_a

      result = {}

      # find/recurse the isa/part_of tree, this gets us a unique array of pertinent Ontologies
      rels = self.tree_relatives(opt.merge(:relationship_type => [opt[:rel1], opt[:rel2]]))

      return {} if rels == nil

      # get all the part_of resuls, anything below and with a part_of, is part_of (since all others are is_a)
      to_nuke = [] # index of rels to delete before next itteration
      rels.each do |r|
        nuke = false # boolean check for delete
        if r.object_relationship_id == opt[:rel2]
          nuke = true
          if opt[:direction] == :parents
            first_result.push(r.ontology_class2) 
          else
            first_result.push(r.ontology_class1) 
          end
        end
        to_nuke.push(r) if nuke 
      end

      # !! don't do uniq on first result

      # try to invoke some loop speedup by deleting values we don't need to loop through  
      rels.delete_if{|r| to_nuke.include?(r)} 
      rels.delete_if{|r| !r.object_relationship_id == opt[:rel1]} # we only need to deal with isas of rel1 now

      # for all of the part_of results also get the is_a children (or whatever rel2 -> rel1 relationship is)

      rels.each do |rel|
        first_result.uniq.each_with_index do |r,i|
          if opt[:direction] == :parents
            second_result.insert(-1, rel.ontology_class2) if (rel.ontology_class1 == r) 
          else
            second_result.insert(-1, rel.ontology_class1) if (rel.ontology_class2 == r)
          end
        end
      end 

      second_result.uniq! # don't imply redundancies from hitting an is_a twice (is this right?! - gives "correct" result AFAIKT)

      (first_result + second_result).each do |r|
        result.merge!(r => (!result.keys.include?(r) ? false : true)) 
      end

      result
  end

  # returns an array of OntologyRelationships 
  # return [] at depths of > 50, this could be incremented or the check changed to properly detect recursion 
  def tree_relatives(options = {})
    opt = {                         # not @opt
      :relationship_type => 'all',  # 'all' or an array of [ObjecRelationship#id, ObjecRelationship#id2 ... ]
      :result => [],
      :depth => 0,                  # ???? can't use :depth because it's tied to the other recurser
      :max_depth => 999,
      :direction => :children       # or :parents
    }.merge!(options.symbolize_keys)
    os = []

    return [] if opt[:depth] > 50 # abort, there is very likely a recursion error

    if opt[:direction] == :children
      cond = 'ontology_class2_id'
    else # else find parents
      cond = 'ontology_class1_id'
    end

    if opt[:depth] < opt[:max_depth] 
      opt[:depth] = opt[:depth] + 1
      if (opt[:relationship_type].to_s == 'all')
        os = OntologyRelationship.find(:all, :include => [:ontology_class1, :object_relationship, :ontology_class2], :conditions => "#{cond} = #{self.id}")
      else
        sql = " (" + opt[:relationship_type].collect{|i| "object_relationship_id = #{i}"}.join(" OR ") + ") "
        os = OntologyRelationship.find(:all, :include => [:ontology_class1, :object_relationship, :ontology_class2], :conditions => "#{cond} = #{self.id} AND #{sql}")
      end

      opt[:result] += os
      os.each do |o|
        if opt[:direction] == :children
          opt[:result] = opt[:result] + o.ontology_class1.tree_relatives(opt.merge!(:result => []))
        else
          opt[:result] = opt[:result] + o.ontology_class2.tree_relatives(opt.merge!(:result => []))
        end
      end
    end

    opt[:result].uniq # redudant recursion is eliminated here?! #.sort{|x,y| x.part1.name <=> y.part1.name}
  end

  def self.auto_complete_search_result(params = {})
    tag_id_str = params[:tag_id]
    return false if (tag_id_str == nil  || params[:proj_id].blank?)

    value = params[tag_id_str.to_sym].split.join('%') # hmm... perhaps should make this order-independent

    lim = case params[tag_id_str.to_sym].length
          when 1..2 then 3 
          when 3..4 then 5 
          else lim = false # no limits
          end 
    result = []

    # order our results
    result += OntologyClass.find(:all, :conditions => ["(labels.name = ? OR ontology_classes.id = ? OR ontology_classes.dbxref = ?) AND ontology_classes.proj_id = ?", value, value.gsub(/\%/, ""), value, params[:proj_id] ], :limit => 1, :include => [:labels] )
    # make this second a regex that does *NOT* match exact label
    result += OntologyClass.find(:all, :conditions => ["(labels.name LIKE ?) AND ontology_classes.proj_id = ?", "%#{value}%", params[:proj_id] ], :limit => lim, :include => [:labels], :order => 'length(labels.name)' )
    result += OntologyClass.find(:all, :conditions => ["(definition LIKE ?) AND ontology_classes.proj_id = ?", "%#{value}%", params[:proj_id] ], :limit => lim, :include => [:labels] )
    result.uniq
  end

  # returns a String in the form of a flat js hash
  def js_flat_hash(options = {})
    @opt = {
      :max_depth => 999,
      :depth => 0,
      :children => [],
      :relationship_type => 'all' # or an Isa#id
    }.merge!(options.symbolize_keys)
    @opt[:depth] = @opt[:depth] + 1
    if @opt[:depth] < @opt[:max_depth]          
      self.tree_child(@opt).each do |n|
        @opt[:children] << n
        n.ontology_class1.js_flat_hash(@opt)   
      end
    end
    return @opt[:children]
  end

  # returns a String in the form of a nested js hash
  def js_hash(options = {})
    @opt = {
      :max_depth => 999,
      :depth => 0,
      :string => '',
      :key_is_id => true, 
      :relationship_type => 'all' # or an Isa#id
    }.merge!(options.symbolize_keys)
    @opt[:depth] = @opt[:depth] + 1
    if @opt[:key_is_id]    
      @n = self.id
    else 
      @n = self.display_name.gsub(/[^a-zA-Z]/, "_") # these all have to be javascript variables, and then can't be used to round trip search, for things like "-"
    end

    if @opt[:depth] < @opt[:max_depth]          
      children = self.tree_child(@opt) 

      if children.size == 0
        @opt[:string] << "#{@n}," # fake size this for the protovis stuff like "#{@n}:10," 
        return @opt[:string] 
      else
        if @opt[:depth] + 1 < @opt[:max_depth]
          @opt[:string] << "#{@n}:{"
        else
          @opt[:string] << "#{@n}:10,"
        end
      end

      children.each do |n|
        n.ontology_class1.js_hash(@opt)   
      end

      if @opt[:depth] + 1 < @opt[:max_depth]
        if children.size > 0
          @opt[:string] << "},"
        else 
          @opt[:string] << "}"
        end
      end
    end

    # hack, to deal with the extra "," in the "join" 
    return "{#{@opt[:string]}}".gsub(/\,\}/, "}") 
  end

  def js_hash2(options = {})
    @opt = {
      :max_depth => 999,
      :depth => 0,
      :string => '',
      :key_is_id => true, 
      :relationship_type => 'all' # or an Isa#id
    }.merge!(options.symbolize_keys)
    @opt[:depth] = @opt[:depth] + 1

    if @opt[:key_is_id]    
      @n = self.id
    else 
      @n = self.display_name.gsub(/[^a-zA-Z]/, "_") # these all have to be JS variables, and then can't be used to round trip search, for things like "-"
    end

    if @opt[:depth] < @opt[:max_depth]          
      children = self.tree_child(@opt) 

      if children.size == 0
        @opt[:string] << "#{@n}:10,"
        return @opt[:string] 
      else
        if @opt[:depth] + 1 < @opt[:max_depth]
          @opt[:string] << "#{@n}:{"
        else
          @opt[:string] << "#{@n}:10," # this is a fake size 
        end
      end

      children.each do |n|
        n.ontology_class1.js_hash2(@opt)   
      end

      if @opt[:depth] + 1 < @opt[:max_depth]
        if children.size > 0
          @opt[:string] << "},"
        else 
          @opt[:string] << "}"
        end
      end
    end

    # hack, to deal with the extra "," in the "join" 
    return "{#{@opt[:string]}}".gsub(/\,\}/, "}").gsub(/\,/, "")
  end

  # TODO: rename to clarify, this is all relationships in which self is present
  def children
    OntologyRelationship.find(:all, :conditions => "((ontology_class1_id = #{self.id}) OR (ontology_class2_id = #{self.id}))", :include => [:ontology_class2]).collect{|o| o.ontology_class2}
  end

  #  used in RelationBrowser
  def direct_children
    OntologyRelationship.find(:all, :conditions => "((ontology_class1_id = #{self.id}) OR (ontology_class2_id = #{self.id}))", :include => [:ontology_class1]).collect{|o| o.ontology_class1}
  end

  def relationships
    OntologyRelationship.find(:all, :conditions => "proj_id = #{self.proj_id} AND ((ontology_class1_id = #{self.id}) OR (ontology_class2_id = #{self.id}))", :include => [:object_relationship] )
  end

  def self.generate_xrefs(options = {})
    opt = {
      :use_mx_ids => false,
      :proj_id => nil,
      :prefix => nil,
      :initial_value => 1,
      :padding => 7, 
      :ontology_classes => [] # OntologyClasses to be numbered
    }.merge!(options.symbolize_keys)  

    return false if !opt[:proj_id] || !opt[:prefix] || opt[:ontology_classes].size == 0
    i = opt[:initial_value].to_i

    OntologyClass.transaction do 
      begin
        opt[:ontology_classes].each do |oc|
          if oc.dbxref.blank?
            if opt[:use_mx_ids]
              oc.dbxref = "#{opt[:prefix]}:#{oc.id.to_s}" 
            else 
              oc.dbxref = "#{opt[:prefix]}:#{i.to_s.rjust(opt[:padding], "0")}" 
              while !oc.valid? 
                i += 1 
                oc.dbxref = "#{opt[:prefix]}:#{i.to_s.rjust(opt[:padding], "0")}" 
              end 
            end    
            oc.save
          end
        end 
      rescue ActiveRecord::RecordInvalid => e
        raise e 
      end
    end
    true 
  end




  # requires :ontology_classes => [] and :proj_id
  def self.strip_candidacy_tags(options = {})
    opt = {
      :ontology_classes => [],
      :proj_id => nil 
    }.merge!(options.symbolize_keys)  
    return false if opt[:proj_id] == nil 
    proj = Proj.find(opt[:proj_id]) 
    return false if proj.ontology_inclusion_keyword.blank?
    begin
      OntologyClass.transaction do
        keywords_to_strip = proj.keywords.tagged_with_keyword(proj.ontology_inclusion_keyword)
        ocs = OntologyClass.tagged_with_keywords(:keywords => keywords_to_strip, :proj_id => opt[:proj_id])
        ocs.each do |p|
          p.tags.each do |t|
            t.destroy if keywords_to_strip.include?(t.keyword)
          end
        end
      end
    rescue
      return false
    end
  end

  # pass params from OntologyController#proofer_batch_create and merge proj_id => id 
  # TODO: revise 
  def self.proofer_batch_create(params)
    raise "method needs revision after class/part split" 
    begin
      @proj = Proj.find(params[:proj_id])
      raise if !@proj 
      raise if params[:part].blank? 

      @count = 0
      params[:taxon_name_id] = params[:term][:taxon_name_id] if params[:term] && !params[:term][:taxon_name_id].blank? # handles batch loading from Proofer
      params[:ref_id] = params[:term][:ref_id] if params[:term] && !params[:term][:ref_id].blank? # handles batch loading from Proofer

      @tn = TaxonName.find(params[:taxon_name_id]) unless params[:taxon_name_id].blank?
      @ref = Ref.find(params[:ref_id]) unless params[:ref_id].blank?
      @ontology_class_for_object_relationship = Part.find(params[:ontology_class_for_object_relationship_id]) unless params[:ontology_class_for_object_relationship_id].blank?
      @object_relationship = ObjectRelationship.find(params[:object_realtionship_id]) unless params[:object_relationship_id].blank?

      ObjectRelationship.transaction do
        params[:part].keys.each do |p|

          # TODO: new Term model, stats etc.
          te = TermExclusion.find_or_create_by_name_and_proj_id(params[:part][p], @proj.id) # BACKGROUND STATS ONLY

          if params[:check][p]
            break if Part.find_by_name_and_proj_id(params[:part][p], @proj.id)

            prt = Part.new(:name => params[:part][p])
            prt.obo_dbxref = params[:dbxref][p] if params[:dbxref] && params[:dbxref][p]
            prt.description = params[:description][p] if params[:description][p]
            prt.taxon_name = @tn if @tn
            prt.ref = @ref if @ref
            prt.save!

            if @isa && @part_for_isa
              @relationship = Ontology.new(:part1_id => prt.id, :part2_id => @part_for_isa.id, :isa_id => @isa.id )
              @relationship.save!
            end

            # add the tag here
            if !params[:tag].blank? && !params[:tag][:keyword_id].blank?
              tag = Tag.new(:keyword_id => params[:tag][:keyword_id], :addressable_type => 'Part', :addressable_id => prt.id)
              tag.notes = params[:tag][:notes] if !params[:tag][:notes].blank?
              tag.referenced_object = params[:tag][:referenced_object] if !params[:tag][:referenced_object].blank?
              tag.save!
            end

            @count += 1

            te.destroy # BACKGROUND STATS - we've used this term now, so it should be reset
          else # BACKGROUND STATS ONLY 
            te.update_attributes(:count => te.count + 1) 
          end
        end
      end

    rescue Exception => e
      raise "#{e} on #{params[:part][p]}"
    end

    return @count 
  end

  ## code below is from deprecated Part that needs complete rewrite

  # TODO: logic is suboptimal, should use a gem engine for param combinations such 
  # needs complete rewrite
  def self.param_search(params)
    raise "param_search not updated since deprecation"
    terms = []
    order = "ordered_by_#{params[:sort_order]}"
    @proj = Proj.find(params[:proj_id])
    if params[:edited]
      if params[:without_relationships] 
        terms = @proj.parts.without_relationships.with_description_status(params[:definition]).with_dbxref_status(params[:dbxref]).changed_by(params[:person_id]).recently_changed(params[:time_ago].to_i.weeks.ago).send(order)
      else
        terms = @proj.parts.with_description_status(params[:definition]).with_dbxref_status(params[:dbxref]).changed_by(params[:person_id]).recently_changed(params[:time_ago].to_i.weeks.ago).send(order)
      end
    else
      if
        terms = @proj.parts.without_relationships.with_description_status(params[:definition]).with_dbxref_status(params[:dbxref]).not_changed_by(params[:person_id]).recently_changed(params[:time_ago].to_i.weeks.ago).send(order)
      else 
        terms = @proj.parts.with_description_status(params[:definition]).with_dbxref_status(params[:dbxref]).not_changed_by(params[:person_id]).recently_changed(params[:time_ago].to_i.weeks.ago).send(order)
      end 
    end
    terms 
  end

  # maintainence
  
  def self.update_all_sensu_positions(proj_id)
    begin
      OntologyClass.transaction do 
        Proj.find(proj_id).ontology_classes.each do |oc|
          puts "updating #{oc.id}"
          oc.update_sensu_positions
        end
      end
    rescue
        "barf"
    end
  end

  def update_sensu_positions 
    sensus.ordered_by_position.each_with_index do |s, i|
      puts "#{s.id} #{i}"
      s.update_attributes(:position => i) 
    end
  end

  protected

  def ensure_that_labels_contains_obo_label
    if !self.obo_label_id.blank? && !self.labels.map(&:id).include?(self.obo_label_id)
      self.sensus << Sensu.create!(:label => Label.find(self.obo_label_id), :ontology_class => self, :ref_id => self.written_by_ref_id) 
    end
    true
  end


end
