class OntologyRelationship < ActiveRecord::Base
  require 'obo_parser'
  has_standard_fields
  include ModelExtensions::Taggable
  include ModelExtensions::DefaultNamedScopes
  include ModelExtensions::MiscMethods

  OBO_TYPEDEFS = ['is_a', 'disjoint_from', 'instance_of', 'inverse_of', 'union_of', 'intersection_of']

  belongs_to :ontology_class1, :foreign_key => 'ontology_class1_id', :class_name => 'OntologyClass'
  belongs_to :ontology_class2, :foreign_key => 'ontology_class2_id', :class_name => 'OntologyClass'
  belongs_to :object_relationship

  named_scope :by_object_relationship, lambda {|*args| {:conditions => ["object_relationship_id = ?",  (args.first || -1)]}}
  named_scope :where_both_ontology_classes_have_dbxrefs, :joins => 'JOIN (ontology_classes oc1, ontology_classes oc2) ON (oc1.id = ontology_class1_id AND oc2.id = ontology_class2_id)', :conditions => 'oc1.dbxref is not null and oc1.dbxref != "" AND oc2.dbxref is not null and oc2.dbxref != ""' 

  validates_presence_of :ontology_class1, :ontology_class2, :object_relationship

  def validate_on_create
    if OntologyRelationship.find_by_ontology_class1_id_and_ontology_class2_id_and_object_relationship_id(ontology_class1_id, ontology_class2_id, object_relationship_id)
      errors.add("The class relationship combination already exists, it therefor ")
    end
  end

  def display_name(options = {})
    ontology_class1.display_name + " " + object_relationship.display_name + " " + ontology_class2.display_name
  end

  # summarizes reference use in the Ontology
  # returns a Hash of {:key => [[Refs] ...] 
  def self.tied_references(options = {})
    return [] if !options[:proj_id]
    proj = Proj.find(options[:proj_id]) 
    result = { }

    result[:used_in_sensus] = proj.refs.used_in_sensus
    result[:used_in_ontology_class_written_by] = proj.refs.used_in_ontology_class_written_by

    result[:used_on_tags_on_classes] = proj.ontology_classes.collect{|p| p.tags.collect{|t| (t.ref_id.blank? ? nil : t.ref)}}.flatten.compact.uniq.sort{|a,b| a.display_name <=> b.display_name}
    result[:used_on_tags_on_labels] =  proj.labels.collect{|p| p.tags.collect{|t| (t.ref_id.blank? ? nil : t.ref)}}.flatten.compact.uniq.sort{|a,b| a.display_name <=> b.display_name}
    result[:used_on_tags_on_sensus] =  proj.sensus.collect{|p| p.tags.collect{|t| (t.ref_id.blank? ? nil : t.ref)}}.flatten.compact.uniq.sort{|a,b| a.display_name <=> b.display_name}

    result[:all] = result.keys.collect{|k| @result[k]}.flatten.compact.uniq.sort{|a,b| a.display_name <=> b.display_name } 

    result
  end

  ## TODO: EVERYTHING BELOW SHOULD BE MOVED 

  # pass a Person
  # TODO: move to `include Roles` , or extend to all ActiveRecord
  def created_or_admin(person)
    if person.id.to_i == self.creator_id.to_i || person.is_admin
      true
    else
      false
    end
  end

  # TODO: move to visualization suite
  # experimental, returns a .dot graph 
  def self.dot(options = {})
    g = GraphViz.new( :G, :type => :digraph )

    Proj.find(options[:proj_id]).ontology_relationships(:include => [:ontology_class1, :ontology_class2, :object_relationship]).each do |o|
      one = g.add_node( o.ontology_class1_id.to_s, "label" => o.ontology_class1.labels.first.display_name )
      two = g.add_node( o.ontology_class2_id.to_s, "label" => o.ontology_class2.labels.first.display_name )

      # Create an edge between the two nodes
      g.add_edge( one, two, "label" => o.object_relationship.interaction, "color" => (o.object_relationship.interaction == "is_a" ? "red" : "blue")  )
    end
    g   
  end

end
