# == Schema Information
# Schema version: 20090930163041
#
# Table name: specimens
#
#  id                 :integer(4)      not null, primary key
#  ce_id              :integer(4)
#  temp_ce            :text
#  parent_specimen_id :integer(4)
#  repository_id      :integer(4)
#  dna_usable         :boolean(1)
#  notes              :text
#  sex                :string(64)
#  stage              :string(64)
#  proj_id            :integer(4)      not null
#  creator_id         :integer(4)      not null
#  updator_id         :integer(4)      not null
#  updated_on         :timestamp       not null
#  created_on         :timestamp       not null
#  preparations       :text
#  disposition        :string(255)
#

class Specimen < ActiveRecord::Base
  has_standard_fields
  set_table_name('specimens')

  include ModelExtensions::Taggable
  include ModelExtensions::Figurable
  include ModelExtensions::DefaultNamedScopes
  include ModelExtensions::MiscMethods

  has_one :most_recent_determination, :class_name => 'SpecimenDetermination', :order => 'det_on DESC, created_on DESC'
  
  belongs_to :ce 
  belongs_to :image_description 
  belongs_to :parent_specimen, :class_name => "Specimen", :foreign_key => "parent_specimen_id" # may need to adjust
  belongs_to :repository
 
  has_many :extracts, :dependent => :destroy
  has_many :measurements, :dependent => :destroy
  has_many :otus, :through => :specimen_determinations, :source => :otu
  has_many :specimens, :class_name => "Specimen", :foreign_key => "parent_specimen_id"
  has_many :specimen_identifiers, :dependent => :delete_all, :before_add => :validate_identifier, :order => 'specimen_identifiers.position'
  has_many :specimen_determinations, :dependent => :delete_all, :before_add => :validate_determination
  has_many :seqs, :dependent => :nullify 
  has_many :type_assignments, :through => :type_specimens, :source => :taxon_name
  has_many :type_specimens, :dependent => :delete_all
  
  named_scope :with_usable_dna, :conditions => {:dna_usable => true} 
  named_scope :determined_as_otu, lambda {|*args| {:conditions => ["specimens.id IN (SELECT specimen_id from specimen_determinations WHERE specimen_determinations.otu_id = ?)", (args.first || -1)] }}
  named_scope :with_type_status, lambda {|*args| {:conditions => ["specimens.id IN (SELECT specimen_id from type_specimens WHERE type_type = ?)", (args.first || -1)] }}
  named_scope :with_type_assignment_for_taxon_name, lambda {|*args| {:conditions => ["specimens.id IN (SELECT specimen_id from type_specimens WHERE taxon_name_id = ?)", (args.first || -1)] }}

  # thanks to folks at Stack Exchange for help with this
  named_scope :with_current_determination, lambda {|*args| 
    { :joins => "JOIN (specimen_determinations sda, otus) " +
        "ON (sda.specimen_id = specimens.id AND sda.otu_id = otus.id) " +
        "LEFT JOIN specimen_determinations sdb " +
        "ON (sdb.specimen_id = specimens.id AND sda.det_on < sdb.det_on)", 
      :conditions => ["sdb.id IS NULL AND otus.id = ?", (args.first || -1)]
    }
  }

  named_scope :with_identifier, lambda {|*args| {:include => :specimen_identifiers, :conditions => ["specimen_identifiers.cached_display_name = ?", (args.first || -1)] }} # exact match
  named_scope :without_type_assignment, {:conditions =>  "specimens.id NOT IN (select specimen_id from type_specimens)"}

  def validate_identifier(identifier)
    if identifier.namespace.blank? or identifier.identifier.blank?
      errors.add_to_base("Specimen identifier needs both an identifier and a namespace.")
      raise ActiveRecord::RecordInvalid, self
    end

   if SpecimenIdentifier.find(:first, :conditions => "identifier = '#{identifier.identifier}' AND namespace_id = #{identifier.namespace_id}")
     errors.add_to_base("The provided identifier is duplicated elswhere in the database.  If you need to use non-unique identifiers consider using tags.") 
     raise ActiveRecord::RecordInvalid, self
   end
  end

  def validate_determination(determination)
     if !determination.otu && determination.name.blank?
        errors.add_to_base("Specimen determination needs a name or OTU.")
        raise ActiveRecord::RecordInvalid, self
     end
  end

  after_update :save_measurements

  def measurement_attributes=(measurement_attributes)
    measurement_attributes.each do |attributes|
      next if attributes == {} 
      if attributes[:id].blank? 
        measurements.build(attributes)
      else
        a = measurements.detect { |t| t.id == attributes[:id].to_i }
        a.attributes = attributes
      end
    end
    true
  end
 
  def save_measurements
    measurements.each do |a|
      if a.measurement.blank?  # measurements of 0 are allowed!!
        a.destroy 
      else
        a.save(true) # passing false ignores validation -- ugh!
      end
    end
  end


  def measurement_for(options = {})
    
    opt = {
      :units => nil,
      :standard_view_id => nil,
      :conversion_factor => nil
    }.merge!(options)

    return nil if opt[:standard_view_id].nil?
    opt.collect{|o| opt.delete(o) if opt[o].nil?}


    Measurement.find(:first, :conditions => opt.merge(:specimen_id => self.id))
  end


  # need to validate that parent isn't part of ancestors

  # TODO: deprecate
  def display_identifiers
    ids = self.specimen_identifiers
    return 'no identifiers' if ids.empty?
    ids.map{|i| "#{i.cached_display_name}"}.join("; ")  
  end
 
  # TODO: deprecate 
  def display_determinations
    @dets = SpecimenDetermination.find_all_by_specimen_id_and_current_det(id, true) 
    return '(no dets)' if @dets.empty?
    @strs = @dets.map{|i| i.display_name }
    @strs.join("; ") 
  end


  def display_name(options = {})
     opt = {
      :type => :identifiers, # :list, :head, :select, :selected, :sub_select, :material_examined_verbose
      :otu => nil,
      :taxon_name => nil,
      :target => ""
     }.merge!(options.symbolize_keys)

    xml = Builder::XmlMarkup.new(:indent=> 2, :target => opt[:target])

    case opt[:type]
    when :identifiers
      if specimen_identifiers.size > 0
        xml << display_identifiers 
      else
        xml << "mx_id: #{self.id}"
      end
    when :select 
      if display_identifiers == 'mxid: #{id.to_s}'
        xml << ''
      else
        xml << "mxid:#{id.to_s}; #{display_identifiers} / "
      end

      xml <<  "("
      if self.most_recent_determination
        xml << self.most_recent_determination.display_for_select 
      else
        xml.i("no dets")
      end 
      xml << ")" 

    when :material_examined_verbose 
      if opt[:otu]
        xml << "#{self.type_status_by_otu(opt[:otu]).collect{|ts| ts.downcase.capitalize}.join(', POSSIBLE ERROR, ASSIGNED TYPE STATUS ADDITIONALLY: ')}" 
      elsif opt[:taxon_name]
        xml << "#{self.type_assignments.with_taxon_name_assignment(opt[:taxon_name]).collect{|ts| ts.type_type.downcase.capitalize}}"  # review
      end
      xml << " #{self.sex.blank? ? "SEX NOT PROVIDED" : self.sex}: #{self.ce.geog.country.name.upcase}: "
      xml << [(self.ce.blank? ? "NO COLLECTING EVENT PROVIDED" :  self.ce.display_name(:type => :material_examined_verbose)), "#{self.specimen_identifiers.collect{|si| si.cached_display_name}.join(",")} (deposited in #{self.repository.coden})"].compact.join(", ")
      xml <<  "."
    when :taxon # returns a string 
      if self.most_recent_determination
        xml << self.most_recent_determination.display_name(:type => :selected)
      else
        xml << "no determinations"
      end 
    else
      xml.div("class" => "dn_#{opt[:type].to_s}") do |i|
        xml.div("id: #{id}", "class" => "dnsid")
        xml.span("specimen: ", "class" => "hd") if opt[:type] == :sub_select # should just css display:none 
        xml.div(self.most_recent_determination.display_name, "class" => "small_grey", "style" => "float:right") if self.specimen_determinations.count > 0 # count or size?  size is saved, count is not
        xml.div(display_identifiers, "class" => "small_grey") if self.specimen_identifiers.size > 0
        xml << self.ce.display_name(:type => :sub_select) if self.ce 
      end 
    end
    opt[:target] 
  end
  
  # TODO: deprecate 
  def display_for_select
    (display_identifiers == 'id: #{id.to_s}' ? '' : " id:#{id.to_s} #{display_identifiers} / ") + "(#{(self.most_recent_determination ? self.most_recent_determination.display_for_select : "<i>no dets</i>")})"
  end

  # TODO: deprecate 
  def display_for_list
    (display_identifiers == 'id: #{id.to_s}' ? '' : " id:#{id.to_s} #{display_identifiers} / ") + "(#{(self.most_recent_determination ? self.most_recent_determination.display_name : "no dets")})"
  end
 
  # TODO: deprecate 
  def display_ce_for_list # uses ce or temp ce depending on which is there
      return self.ce.display_for_list if self.ce
      return self.temp_ce[0..25] if self.temp_ce
  end

  # TODO: deprecate
  def display_ce
     return self.ce.display_name if self.ce
     return self.temp_ce if self.temp_ce
  end

  # TODO constantize
  def self.type_types
    ['holotype', 'paratype', 'syntype', 'lectotype', 'neotype', 'paralectotype']
  end
 
  def mappable
    return true if self.ce && self.ce.mappable
    false
  end
    
  # redundant with above?
  def has_lat_long
    if self.ce and !self.ce.latitude.blank? and !self.ce.longitude.blank?
      return true
    end
    false
  end

  # TOTO: named scope
  # def self.by_identifier(proj_id)
  #  return Specimen.find_by_sql(["SELECT specimens.*, specimen_identifiers.identifier FROM specimen_identifiers 
  #  RIGHT JOIN specimens ON specimen_identifiers.specimen_id = specimens.id WHERE ( (specimens.proj_id) = (?)) ORDER BY specimen_identifiers.identifier;", proj_id ] )
  # end

  # returns this specimens type status string, or false, for the given Otu
  # Rules- if Otu#taxon_name.blank? then all statuses NOT tied to a taxon name are returned (Pragmatically there should only be one pre-press assignment, otherwise you've got some messed up decision making)
  #        if Otu#taxon_name then grab the status for that TaxonName
  def type_status_by_otu(otu)
    if otu.taxon_name.blank?
      return self.type_specimens.without_taxon_name_assignment.collect{|t| t.type_type}.join(",") if self.type_assignments.count > 0
    else
      return self.type_specimens.with_taxon_name_assignment(otu.taxon_name.id).collect{|t| t.type_type}.join(",") if self.type_assignments.count > 0
    end
    return false
  end

  # flexible Specimen and Lot importing, with lots of verification built in
  # only :file, :proj_id are required
  # see also approaches by the Nearctic Spider Database and the HOL databases OSU
  def self.read_from_batch(options = {})
    @opt = {
      :expand_line_breaks => true,
      :line_break_symbol => '||',
      :new_label_symbol =>  '++',
      :col_sep => "\t",
      :proj_id => nil,
      :save => false
    }.merge!(options)

    raise Specimen::SpecimenBatchParseError, "No file provided"  if @opt[:file].blank?
    raise Specimen::SpecimenBatchParseError, "No proj_id provided"  if @opt[:proj_id].blank?

    # not used for verification yet
    # this is the master list of possibly parsed fields, see the wiki for explanation
    @legal_headers = %w(
      ce_id
      count
      data_entry_by
      det_basis
      det_name
      det_otu_id  
      det_year 
      determiner
      identifier
      notes
      otu_id
      otu_name 
      repository_coden
      repository_id
      sex
      stage
      taxon_name_id
      taxon_name_string
      type
      type_of_taxon_name_id
      verbatim_label
      latitude  
      longitude
      elev_min
      elev_max
      geog_id
      method
      locality 
      sd_d
      sd_m
      sd_y
      ed_d
      ed_m
      ed_y
   ).collect{|h| h.to_sym}
    # possible additions - Tags, expanding ces

    @results = {:specimens => [], :lots => [], :ces => []}
   
    incoming_encoding = CMess::GuessEncoding::Automatic.guess(@opt[:file])
    translated_file = Iconv.iconv('UTF-8', incoming_encoding, @opt[:file]).first

    # move to another method?
    case incoming_encoding
    when "UTF-16LE"
      translated_file = translated_file[3..translated_file.size] # KLUDGE - hacked to work, to broadly tested
    else
    end

    recs = FasterCSV.parse(translated_file, :headers => true, :row_sep => :auto, :header_converters => nil, :col_sep => @opt[:col_sep], :encoding => 'u') # reading from a string http://fastercsv.rubyforge.org/
   
    # first deal with validation
    # do all the checks that don't require database hits  
    identifiers = {} 
    recs.each_with_index do |r,x|
      i = x+1
      raise Specimen::SpecimenBatchParseError, "Data row #{i} has both neither count (Lot) nor identifier (Specimen), add one." if r.fields('count').first.blank? && r.fields('identifier').first.blank?
      #  raise Specimen::SpecimenBatchParseError, "Data row #{i} has both count (Lot) and identifier (Specimen), remove one." if !r.fields('count').first.blank? && !r.fields('identifier').first.blank?
      raise Specimen::SpecimenBatchParseError, "Data row #{i} has no taxon or otu reference" if r.fields('otu_id').first.blank? && r.fields('otu_name').first.blank? && r.fields('taxon_name_string').first.blank? && r.fields('taxon_name_id').first.blank?
      raise Specimen::SpecimenBatchParseError, "Data row #{i} has both otu and taxon_name references, remove one." if (!r.fields('otu_id').first.blank? || !r.fields('otu_name').first.blank?) && (!r.fields('taxon_name_id').first.blank? || !r.fields('taxon_name_string').first.blank?)
      raise Specimen::SpecimenBatchParseError, "Data row #{i} has both ce_id and verbatim label, remove one." if (!r.fields('ce_id').first.blank? && !r.fields('verbatim_label').first.blank?)
      raise Specimen::SpecimenBatchParseError, "Data row #{i} has both repository_id and repository_coden, remove one." if !r.fields('repository_coden').first.blank? && !r.fields('repository_id').first.blank?
      raise Specimen::SpecimenBatchParseError, "Data row #{i} has not enough information to define type status." if (!r.fields('type').first.blank? || !r.fields('type_of_taxon_name_id').first.blank?) && (r.fields('type').first.blank? || r.fields('type_of_taxon_name_id').first.blank?) 
      raise Specimen::SpecimenBatchParseError, "Data row #{i} has an invalid count." if !r.fields('count').first.blank? && !(r.fields('count').first.to_i > 1)
      raise Specimen::SpecimenBatchParseError, "Data row #{i} has both det_otu_id and det_name, remove one." if !r.fields('det_otu_id').first.blank? && !r.fields('det_name').first.blank?
      raise Specimen::SpecimenBatchParseError, "Data row #{i} has valid count but no otu_id." if !r.fields('count').first.blank? && (r.fields('count').first.to_i > 1) && r.fields('otu_id').first.blank?

      raise Specimen::SpecimenBatchParseError, "Data row #{i} has an identifier repeated in row #{identifiers[r.fields('identifier').first]}." if identifiers[r.fields('identifier').first]

      identifiers.merge!(r.fields('identifier').first => i) if !r.fields('identifier').first.blank?
      # validate additional determination
    end   

    # reloop and instantiate Specimens/Lots/Ces etc.
   
    @ces = {} # verbatim_label => Ce (index duplicates so we don't create them 2x)
    begin
     Specimen.transaction do 
       recs.each_with_index do |r,x|
        i = x + 1
        
        # find the Otu reference
        # via reference to TaxonName
        if r.fields('otu_id').first.blank? && r.fields('otu_name').first.blank? # it's a TaxonName
          @taxon_name = nil
          if !r.fields('taxon_name_id').first.blank? && !r.fields('taxon_name_string').first.blank?
            @taxon_name = TaxonName.find(r.fields('taxon_name_id').first)
            raise Specimen::SpecimenBatchParseError, "Data row #{i} has a taxon_name_id that doesn't match the taxon_name_string." if taxon_name.display_for_list != r.fields('taxon_name_string').first
          elsif !r.fields('taxon_name_id').first.blank?
            @taxon_name = TaxonName.find(r.fields('taxon_name_id').first)
          else # must a taxon_name
            taxon_names = TaxonName.find(:all, :conditions => {:name => r.fields('taxon_name_string').first} ) # make this more complicated, break down names
            raise Specimen::SpecimenBatchParseError, "Data row #{i} has a taxon_name_string that does not match a name in the DB." if taxon_names.blank? || taxon_names.size == 0 
            raise Specimen::SpecimenBatchParseError, "Data row #{i} has a taxon_name_string that matches more than one taxon name." if taxon_names.size > 1
            @taxon_name = taxon_names[0]
          end 

          otus = Otu.find(:all, :conditions => {:taxon_name_id =>  @taxon_name.id, :proj_id =>  @opt[:proj_id]})
          raise Specimen::SpecimenBatchParseError, "Data row #{i} has a taxon_name_id that matches more than one Otu." if otus.size >  1
          raise Specimen::SpecimenBatchParseError, "Data row #{i} has a taxon_name_id that has no matching Otu." if otus.size == 0
          
          @otu = otus[0]

        # via reference to Otu
        else 
          if !r.fields('otu_id').first.blank? && !r.fields('otu_name').first.blank?
            @otu = Otu.find(r.fields('otu_id').first)
            raise Specimen::SpecimenBatchParseError, "Data row #{i} has a otu_id with non-matching Otu name." if @otu.name != r.fields('otu_name').first 
          elsif !r.fields('otu_id').first.blank?
            # throws not found to catch  
            @otu = Otu.find(r.fields('otu_id').first, :conditions => {:proj_id => @opt[:proj_id]})
          else
            otus =  Otu.find(:all, :conditions => {:name => r.fields('otu_name').first, :proj_id => @opt[:proj_id]})
            raise Specimen::SpecimenBatchParseError, "Data row #{i} has no matching Otu for otu_name #{r.fields('otu_name').first}." if otus.nil? || otus.size == 0
            raise Specimen::SpecimenBatchParseError, "Data row #{i} has an otu_name with 2 or more matching names in DB, include an otu_id or resolve in DB." if otus.size > 1
            @otu = otus[0] 
          end
        end 

       # at this point we have an @otu, or we've raised

        # find a Person if requested
        @person = Person.find(:first, :conditions => {:login => r.fields('data_entry_by').first}) if !r.fields('data_entry_by').first.blank? 

        # is it a lot or specimen?
        @specimen = nil
        @lot = nil 

        if r.fields('count').first.blank? # a specimen 
          @specimen = Specimen.new() 
          @specimen.creator = @person if @person
          @specimen.updator = @person if @person

          # have to save before adding manys 
          @specimen.save if @opt[:save]
         
          # we create at least one deterimination
          @specimen.specimen_determinations << SpecimenDetermination.new(:otu => @otu, :det_on => Time.now) 
        else # it's a lot
          @lot = Lot.new()
          @lot.creator = @person if @person
          @lot.updator = @person if @person
          @lot.key_specimens = r.fields('count').first # it has to have this if we are this far
          @lot.otu = @otu                              # has to have this too 
          @lot.save if @opt[:save] 
        end

        # TODO: should lot/specimen identifiers be merged? ... probably
        # handle identifiers
        if !r.fields('identifier').first.blank?
          id_parts = r.fields('identifier').first.split 
          raise Specimen::SpecimenBatchParseError, "Data row #{i} has a malformed identifier (#{r.fields('identifier').first})." if id_parts.size != 2
          @namespace = Namespace.find(:first, :conditions => {:name => id_parts[0]})
          raise Specimen::SpecimenBatchParseError, "Data row #{i} has an identifier whose namespace (#{id_parts[0]}) is not present in the DB." if @namespace.nil? 
          if !@specimen.blank?
            si =  SpecimenIdentifier.new(:namespace => @namespace, :identifier => id_parts[1])
            si.creator = @person if @person
            si.updator = @person if @person
            @specimen.specimen_identifiers << si 
          else
            li = LotIdentifier.new(:namespace => @namespace, :identifier => id_parts[1])
            li.creator = @person if @person
            li.updator = @person if @person
            @lot.lot_identifiers << li
          end
        end

        # handle the repository
        @repository = Repository.find(r.fields('repository_id').first) if !r.fields('repository_id').first.blank?
        @repository = Repository.find(:first, :conditions => {:coden => r.fields('repository_coden').first}) if !r.fields('repository_coden').first.blank?
        raise Specimen::SpecimenBatchParseError, "Data row #{i} has Repository which can not be found." if (!r.fields('repository_coden').first.blank? || !r.fields('repository_id').first.blank?) && @repository.nil?
        @lot.repository = @repository if @repository && @lot
        @specimen.repository = @repository if @repository && @specimen

        if !r.fields('type').first.blank? 
          raise Specimen::SpecimenBatchParseError, "Data row #{i} has an invalid type_of_taxon_name_id (#{r.fields('type_of_taxon_name_id')})." if !TaxonName.find(r.fields('type_of_taxon_name_id').first)
          @specimen.type_specimens << TypeSpecimen.new(:taxon_name_id => r.fields('type_of_taxon_name_id').first, :type_type => r.fields('type').first)
        end 

        # handle determinations
        if !r.fields('det_otu_id').first.blank? || !r.fields('det_name').first.blank?
          sd = SpecimenDetermination.new()
          sd.det_on = Date.new( r.fields('det_year').first.to_i ) if !r.fields('det_year').first.blank? 
          if !r.fields('det_otu_id').first.blank? 
            o = Otu.find(r.fields('det_otu_id').first)
            raise Specimen::SpecimenBatchParseError, "Data row #{i} has a det_otu_id which can not be found." if !o
            sd.otu = o
          else
            sd.name = r.fields('det_name').first  
          end
          sd.determination_basis = r.fields('det_basis').first if !r.fields('det_basis').first.blank?
          sd.creator = @person if @person
          sd.updator = @person if @person
          @specimen.specimen_determinations << sd
        end
  
        # handle collecting events
        if !r.fields('verbatim_label').first.blank? || !r.fields('ce_id').first.blank?
          if !r.fields('verbatim_label').first.blank? 
            l = r.fields('verbatim_label').first.split(@opt[:new_label_symbol]).map{|i| i.strip.split(@opt[:line_break_symbol]).map{|j| j.strip}.join("\n")}.join("\n\n")
            if !@ces[l] 
              @ce = Ce.new(:verbatim_label => l)           
              @ce.latitude = r.fields('latitude').first if !r.fields('latitude').first.blank?
              @ce.longitude = r.fields('longitude').first if !r.fields('longitude').first.blank?
              @ce.elev_min = r.fields('elev_min').first if !r.fields('elev_min').first.blank?
              @ce.elev_max = r.fields('elev_max').first if !r.fields('elev_max').first.blank?
              @ce.method = r.fields('method').first if !r.fields('method').first.blank?
              @ce.locality = r.fields('locality').first if !r.fields('locality').first.blank?

              @ce.sd_d = r.fields('sd_d').first if !r.fields('sd_d').first.blank?
              @ce.sd_m = r.fields('sd_m').first if !r.fields('sd_m').first.blank?
              @ce.sd_y = r.fields('sd_y').first if !r.fields('sd_y').first.blank?

              @ce.ed_d = r.fields('ed_d').first if !r.fields('ed_d').first.blank?
              @ce.ed_m = r.fields('ed_m').first if !r.fields('ed_m').first.blank?
              @ce.ed_y = r.fields('ed_y').first if !r.fields('ed_y').first.blank?

              @ces.merge!(@ce.verbatim_label => @ce) # keep track of what we've added !! IMPORTANT: Uniqueness is determined solely by verbatim label
              
            else
              @ce = @ces[l]
            end
          else
            @ce = Ce.find(r.fields('ce_id').first)
            raise Specimen::SpecimenBatchParseError, "Data row #{i} has ce_id which is not in the DB." if !@ce
          end
          
          @lot.ce = @ce if @lot
          @specimen.ce = @ce if @specimen
        end
        
        ## at this point we have either @specimen OR @lot instantiated or have raised 
        ['sex', 'stage', 'notes'].each do |p|
          if !r.fields(p).first.blank?
            @lot[p] = r.fields(p).first if @lot
            @specimen[p] = r.fields(p).first if @specimen
          end
        end
         
        @results[:specimens].push @specimen if @specimen
        @results[:lots].push @lot if @lot
      end
   
       @results[:ces] = @ces.values

        # save the identifiers/ces etc.
        if @opt[:save]
          [:ces, :specimens, :lots].each do |s|
            @results[s].map{|o| o.save!}
          end
        end
       
      end # end the transaction

    rescue Specimen::SpecimenBatchParseError
      raise 
    rescue ActiveRecord::RecordInvalid => e
      raise "Error during saving a record (#{e})."
    end

    return @results
  end

  # you must have $person_id and $proj_id set
  def make_clone 
    s = self.clone
    s.created_on = Time.now
    s.updated_on = Time.now
    
    s.save
    s.specimen_identifiers.destroy_all

    sd = self.most_recent_determination.clone
    sd.specimen_id = s
    s.specimen_determinations << sd
    s.save
    s
  end

  def self.find_for_auto_complete(value)
    find_by_sql [
      "SELECT DISTINCT s.* FROM specimens s 
      LEFT JOIN specimen_identifiers i ON s.id = i.specimen_id
      LEFT JOIN specimen_determinations sd ON s.id = sd.specimen_id
      LEFT JOIN otus o on o.id = sd.otu_id
      LEFT JOIN taxon_names t ON o.taxon_name_id = t.id 
      LEFT JOIN ces c on s.ce_id = c.id
      LEFT JOIN geogs g on c.geog_id = g.id
      WHERE s.proj_id = #{$proj_id} 
      AND (s.id = ? OR o.name LIKE ? OR c.verbatim_label like ? OR o.matrix_name LIKE ? OR o.manuscript_name LIKE ? OR t.name LIKE ? OR i.cached_display_name LIKE ? or i.identifier = ? OR sd.name LIKE ? OR g.name LIKE ? OR c.verbatim_label like ?) LIMIT 30",
       "#{value}", "#{value.downcase}%", "%#{value.downcase}%", "%#{value.downcase}%", "%#{value.downcase}%", "#{value.downcase}%", "#{value}%", value, "%#{value.downcase}%", "%#{value.downcase}%", "%#{value.downcase}%"]
  end

   class SpecimenBatchParseError < ApplicationController::BatchParseError
   end

end



