class Linker

  # A utility class that matches words in a block of text to Labels/OntologyClasses in mx

  require 'Strings' 

  attr_reader :all_words          # Array of all individual words
  attr_reader :link_root
  attr_reader :proj_id
  attr_reader :public_server_name # Proj#public_server_name
  attr_reader :text_to_link 

  def initialize(options = {}) 
      @opt = {
        :proj_id => nil,                                      # required, the project containing the "ontology" and plural keyword
        :exclude_common_words => false,                       # subtracts COMMON_WORDS from @all_words on init
        :exclude_common_words_of_size_smaller_than => 99999,  # if exclude_common_words is true then only exclude COMMON_WORDS of length < this 
        :minimum_word_size => 1,                              # exclude from @all_words words smaller than this
        :adjacent_words_to_fuse => 0,                         # e.g. given "A B C" and 2 (n-1) then @all_words is [A, A B, B C, A B C]
        :common_words => COMMON_WORDS,                        # COMMON_WORDS is in /lib, can provide custom list
        :public_server_name => nil                            # REQUIRED for linked_text in public to return linked text
      }.merge!(options)
  
    return nil if !@opt[:incoming_text] || (@opt[:incoming_text].size == 0) || @opt[:proj_id].blank?
    # return nil if not @proj = Proj.find(@opt[:proj_id]) # need this for the public server name - move to attribute and find at first call

    if self.reset!(@opt) 
      return self # true
    end
    nil 
  end
  
  def reset!(options = {})
    opt = {}.merge!(options)
    return nil if !opt[:incoming_text] || opt[:incoming_text].empty? 

    # set the accessible vars
    @public_server_name = opt[:public_server_name] 
    @proj_id = opt[:proj_id]
    @all_words =  Strings::word_set_fusing_adjacent(opt)
    @text_to_link = opt[:incoming_text] 
  
     @link_root = opt[:is_public] ? 
                      ( "http://#{@public_server_name}/projects/#{@proj_id.to_s}/public") :
                      ( "/projects/#{@proj_id.to_s}" )

    # exclude common words if requested 
    # we could consider using prefix trees if they would speed things up, see: http://blade.nagaokaut.ac.jp/cgi-bin/scat.rb/ruby/ruby-talk/226837
    if opt[:exclude_common_words]
      # @all_words = @all_words - opt[:common_words] # TODO: benchmark this approach vs. below 
     rgx = Regexp.new(opt[:common_words].collect{|w| "\\b(?:#{w})\\b"}.join("|")) # compile one giant regex to match against, should be faster than double loop
      @all_words.reject!{|w| w =~ rgx}
    end 
       
    true
  end

  # render HTML linkages
  def linked_text(options = {})  
    # options set post initialization
    opt = {
      :is_public => false,
      :popup => false,
      :css_class => 'passed',
      :min_length => 1,
      :expandable_target => false,  # a hidden div that surrouds
      :mode => 'link'    # 'link' or 'bracket', the latter largely for debugging
    }.merge!(options)

    out = ((@text_to_link.nil? || @text_to_link.empty?) ?  '<i>none</i>' : @text_to_link  )  + " " # hack to allow gsub to match end of lines
   
    return out if (opt[:is_public] && !@public_server_name)  # this will likely cause confusion

    popup_txt = (opt[:popup] ? 'target="_blank"' : '') # do we want to the link to popup in a new window

    # generate a Hash of Arrays like "foo" => [1,2] where keys are Label names and values are arrays of OntologyClass#id 
    matchable_words = {}
    self.link_set(opt).each do |l|
      matchable_words.merge!(l.name => OntologyClass.by_proj(@proj_id).by_label_including_plurals(l.name).collect{|oc| oc.id})
    end

    matchable_words.keys.sort{|a, b| b.length <=> a.length}.each do |w| # longest to shortest to prevent nested matches
      if matchable_words[w].size == 1 # a single definition for the word (not a homonym in the ontology)
        if opt[:mode] == 'link' 
          # tricky!  we assume that *ALREADY* linked text is bound in a certain manner, and don't hit it with this regex 
          out.gsub!(/[^\>]\b(?!\>)\B*#{w}(?!\<)\b|^#{w}(?!\b\<)\b/i) {|s| 
          "<notextile>&nbsp;<a class=\"#{opt[:css_class]}\" #{popup_txt} href=\"QQQQ/ontology_class/show/#{matchable_words[w].first.to_s}\">#{w}</a></notextile>" 
          }
        elsif opt[:mode] == 'bracket'
          # nasty regex, but seems to work, note inversion of brackets for href links
          # we have two cases, one is that hits words at the start of a string, the other is everything else
          # (?!\[\b) - look ahead for something not [ and a word boundary
          # [^\[] - a character not a [
          # \A at the begining
          # (?!\b\])\b look ahead for something not ] and a word boundary
          out.gsub!(/[^\[]\b(?!\[)\B*#{w}(?!\])\b|^#{w}(?!\b\])\b/i) {|s| " [#{w}]"} 
        end 

      # TODO this case isn't updated for brackets yet, and might be DEPRECATED
      elsif matchable_words[w].size >  1
        # there is more than one definition (Part that matches) for this word
        # link_text = matchable_words[w].collect{|o| '<notextile>&nbsp;<a class="' + opt[:css_class] + '" ' + popup_txt + ' href="QQQQ/ontology_class/show/' + o.to_s + '">' + o.to_s + '</a></notextile>'}.join(" ") 
        # out.gsub!(/[^\>]\b(?!\>)\B*#{w}(?!\<)\b|^#{w}(?!\b\<)\b/i) {|s|  w + " ["  + link_text + "]"}
        # out.gsub!(/\b(?!\>)\B*#{w}(?!\<)\b|^#{w}(?!\b\<)\b/i) {|s| w + " ["  + link_text + "]" }
        # out.gsub!(/(\A#{w}|[^\>]#{w})(\s*[^\<])/i) {|s| w + " ["  + link_text + "]" + $2 } - old version
      end
    end

    # some cleanup/final checking
    out.gsub!(/QQQQ/, @link_root) if out.size > 0 && opt[:mode] != 'bracket' # gets rid of the problem of the link_root having a defined word in it, rare, but possible
    out.gsub!(/\n/, '<br />') if opt[:mode] == 'link'
    out.strip
  end

# private

  # returns an Array of Labels that are possible links in the text
  # default allows Labels not tied to OntologyClasses (i.e. without definitions)
  def link_set(options = {})
    @opt = {
      :exclude_blank_descriptions => false,
      :include_plural => true,               # also attempt to match on plural form Tags of Parts
      :min_length => 1                       # only include Parts with name greater than min_length
      # may need a max_words cuttoff here 
    }.merge!(options)

    return [] if @all_words == [] || @all_words == nil 

    tmp_words = @all_words.reject{|w| w.length < @opt[:min_length]}

    if @opt[:exclude_blank_descriptions]
     Proj.find(@proj_id).labels.with_definitions.with_label_from_array(tmp_words).ordered_by_label_length.ordered_by_name 
    else
      Proj.find(@proj_id).labels.with_label_from_array(tmp_words).ordered_by_label_length.ordered_by_name 
    end
  end

  def unmatched(options ={})
     @opt = {
       :minimum_word_size => 1 # return only words of size length, can be different than on init
     }.merge!(options)
    return [] if @all_words.size == 0
    matchable_words = self.link_set(@opt).inject([]){|sum, p| sum += p.all_forms}
    # this is word in any, we want exact matches 
    # rgx = Regexp.new( matchable_words.collect{|w| "\\b(?:#{w})\\b"}.join("|") ) # compile one giant regex to match against, should be faster than double loop
    (@all_words - matchable_words).reject{|w| w.length < @opt[:minimum_word_size]}.sort{|a,b| (b.length <=> a.length) && (a <=> b)}
  end

end


