require File.dirname(__FILE__) + '/../test_helper'

class LabelTest < ActiveSupport::TestCase

  def setup
    set_before_filter_vars
    @ref = Ref.create!
  end

  test "valid label forms" do
    l = Label.new(:name => ' invalid preceeding space')
    assert !l.valid?
    l.name = 'invalid postfixed space '
    assert !l.valid?
    l.name = 'valid with space in middle'
    assert l.valid?
  end
    

  test "ontology_classes_for_plurals" do
     l = Label.create!(:name => "a")
     assert !l.has_definition?
     oc = OntologyClass.create!(:definition => "Foo is a thing with foosball.", :written_by => @ref)
     Sensu.create!(:ref => @ref, :ontology_class => oc, :label => l)
     l.reload
     assert l.has_definition?
  end

  test "named_scope_with_label_from_array" do
    %w(cow moose beaver horse e).each do |l|
      Label.create(:name => l)
    end
    assert_equal ["e", "moose"], Label.with_label_from_array(['moose', 'e']).map(&:name).sort
    assert_equal [], Label.with_label_from_array([]).map(&:name).sort
    assert_equal [], Label.with_label_from_array(['elephant']).map(&:name).sort
  end

  test "has_definition?" do
     l = Label.create!(:name => "a")
     assert !l.has_definition?
     oc = OntologyClass.create!(:definition => "Foo is a thing with foosball.", :written_by => @ref)
     Sensu.create!(:ref => @ref, :ontology_class => oc, :label => l)
     l.reload
     assert l.has_definition?
  end

  test "has_plural?" do
    l = Label.create!(:name => "a")
    l1 = Label.create!(:name => 'as', :plural_of_label => l)
    l.reload
    l1.reload
    assert l.has_plural?
    assert !l1.has_plural?
  end

  test "is_plural?" do
    l = Label.create!(:name => "a")
    l1 = Label.create!(:name => 'as', :plural_of_label => l)
    assert !l.is_plural?
    assert l1.is_plural?
  end

  test "that labels are not pluralized multiple times" do
    l = Label.create!(:name => "a")
    l1 = Label.create!(:name => 'as', :plural_of_label => l)
    l2 = Label.new(:name => 'bs', :plural_of_label => l)
    assert !l2.valid?
  end

  test "all_forms" do
    l = Label.create!(:name => "a")
    l1 = Label.create!(:name => 'as', :plural_of_label => l)
    assert_equal %w(a as), l.all_forms
  end

  test "without_ontology_classes_but_used_in_ontology_class_definitions" do
    l = Label.create!(:name => "foo")
    l1 = Label.create!(:name => 'foos')
    oc = OntologyClass.create!(:definition => "Foo is a thing with foosball.", :written_by => @ref)
    h = {l => [oc]}
    assert_equal h, Label.without_ontology_classes_but_used_in_ontology_class_definitions(:proj_id => $proj_id)
  end

  test "all_labels_for_ontology_class" do
    l = Label.create!(:name => "a")
    l1 = Label.create!(:name => 'as', :plural_of_label => l)
        r = Ref.create!
    oc = OntologyClass.create!(:definition => "Foo is a thing with foosball.", :written_by => @ref)

    s = Sensu.create!(:ref => r, :ontology_class => oc, :label => l)
    assert_equal [l,l1], Proj.find($proj_id).labels.all_for_ontology_class(oc)
  end

  test "synonyms_by_ontology_class" do
    l = Label.create!(:name => "a")
    l1 = Label.create!(:name => 'b')
    oc = OntologyClass.create!(:definition => "Foo is a thing with foosball.", :written_by => @ref)
    s = Sensu.create!(:ref => @ref, :ontology_class => oc, :label => l)
    s1 = Sensu.create!(:ref => @ref, :ontology_class => oc, :label => l1)
    l.reload
    l1.reload
    
    assert_equal [l1], l.synonyms_by_ontology_class(oc)
    assert_equal [l], l1.synonyms_by_ontology_class(oc)
  end
  
  
end
