
require File.dirname(__FILE__) + '/../test_helper'

class OntologyClassTest < ActiveSupport::TestCase

  # several of the tests here are mashups of OntologyClass and OntologyRelationship, if in doubt place place/leave these here
  
  def setup
    set_before_filter_vars
    @proj = Proj.find($proj_id) 
    
    @isa = ObjectRelationship.create!(:interaction => 'isa') 
    @partof = ObjectRelationship.create!(:interaction => 'partof') 

    @ref_stub = Ref.create!
  end

  test "fill_blank_xrefs" do
    @o = OntologyClass.create!(:definition => 'Foos and bars', :written_by => @ref_stub)
    assert @o.dbxref.nil?

    assert !OntologyClass.generate_xrefs(:proj_id => @proj.id, :prefix => "Foo") # no :ontology_classes
    assert !OntologyClass.generate_xrefs(:ontology_classes => [@o], :prefix => "Foo") # no :proj_id
    assert !OntologyClass.generate_xrefs(:ontology_classes => [@o], :proj_id => @proj.id) # no :prefix

    OntologyClass.generate_xrefs(:ontology_classes => [@o], :proj_id => @proj.id, :prefix => "Foo")
    @o.reload
    assert_equal "Foo:0000001", @o.dbxref
  end

  test "preferred label logic" do
    @l1 = Label.create!(:name => "foo")
    @l2 = Label.create!(:name => "bar")
    @l3 = Label.create!(:name => "blorf")

    @o = OntologyClass.create!(:definition => 'Foos and bars', :written_by => @ref_stub, :obo_label => @l1)
    
    # a background sensu is created here, reload so another is not found
    @o.reload

    # bind labels to classes 
    @s1 = Sensu.new(:label => @l2, :ref => @ref_stub) 
    @s2 = Sensu.new(:label => @l3, :ref => @ref_stub) 
    
    @o.sensus << @s1 
    @o.save
    @o.sensus << @s2
    @o.save

    @o.reload

    assert_equal 3, @o.labels.size # one for the obo_label, one for each of the sensus

    assert_equal 2, @s1.position 
    assert_equal 3, @s2.position 

    assert_equal 'foo', @o.label_name(:type => :preferred) # foo was the first to be added

    # reverse the order
    @o.sensus.each_with_index do |s,i|
      s.position = 3 - i
      s.save 
    end
    
    @o.reload
    assert_equal 'blorf', @o.label_name(:type => :preferred)

    @s1.destroy
    @s2.destroy
    assert_equal 'foo', @o.label_name(:type => :preferred)

    @o.sensus.destroy_all

    assert_equal 'no labels', @o.label_name(:type => :preferred)
  end

  test "records with dbxrefs can not be destroyed in typical fashion" do
    o = OntologyClass.create!(:definition => 'Foos and bars', :dbxref => "FOO:123", :written_by => @ref_stub)
    assert !o.destroy
    assert o.errors.on(:dbxref) 
  end

  test "setting a obo label id automatically adds to self#labels if not present" do
    l = Label.create!(:name => "foo")
    o = OntologyClass.create!(:definition => 'Foos and bars', :written_by => @ref_stub, :obo_label => l)
    o.reload
    assert_equal [l], o.labels
  end
  
  test "dbxrefs must be in foo:bar format" do
      o = OntologyClass.new(:definition => 'Foos and bars', :written_by => @ref_stub, :dbxref => 'blorf')
      assert !o.valid?
      o.dbxref = " asdf:asdf" # no preceeding spaces
      assert !o.valid?
      o.dbxref = "asdf:asdf " # no postifixed spaces
      assert !o.valid?
      o.dbxref = "asdf:1234" # no postifixed spaces
      assert o.valid?
  end
  
  test "defintion is required" do
    o = OntologyClass.new
    assert !o.valid?
  end

  def setup_for_simple_plural_tests
    @o = OntologyClass.create!(:definition => 'Foos and bars', :written_by => @ref_stub)
    @l1 = Label.create!(:name => "foo")
    @l2 = Label.create!(:name => "foos", :plural_of_label_id => @l1.id)
    @l3 = Label.create!(:name => "bar")
   
    # bind labels to classes 
    @s = Sensu.create!(:label => @l1, :ontology_class => @o, :ref => @ref_stub)
    @s1 = Sensu.create!(:label => @l3, :ontology_class => @o, :ref => @ref_stub)
  end

  test "that all_labels_through_sensus has_many works" do
    setup_for_simple_plural_tests
    @another_ref_stub = Ref.create!
    Sensu.create!(:label => @l1, :ontology_class => @o, :ref => @another_ref_stub)
    
    assert_equal [@l3, @l1, @l1], @o.all_labels_through_sensus.ordered_by_name
  end

  test "that labels_through_sensus has_many returns unique labels" do
    setup_for_simple_plural_tests
    @another_ref_stub = Ref.create!
    Sensu.create!(:label => @l1, :ontology_class => @o, :ref => @another_ref_stub)
    assert_equal [@l3, @l1], @o.labels.ordered_by_name
  end


  test "that plural labels return ontology classes when tied through singular of label in sensu" do
    setup_for_simple_plural_tests
    assert_equal [@o], Proj.find($proj_id).ontology_classes.by_label_including_plurals(@l2.name) # String 
    assert_equal [@o], Proj.find($proj_id).ontology_classes.by_label_including_plurals(["don't match and single quote test", @l2.name]) # Array
    assert_equal [], Proj.find($proj_id).ontology_classes.by_label_including_plurals() # nil

  end

  test "all labels includes plural forms" do
    setup_for_simple_plural_tests
    assert_equal %w/bar foo foos/, @o.all_labels
  end

  test "add IP vote on create" do
     o = OntologyClass.create!(:definition => 'Foos and bars', :illustration_IP_vote => '127.0.0.1', :written_by => @ref_stub)
     assert_equal 1, o.illustration_IP_votes.size
     assert_equal ['127.0.0.1'], o.illustration_IP_votes
  end

  test "add IP vote post create" do
     o = OntologyClass.create!(:definition => 'Foos and bars', :written_by => @ref_stub)
     assert_equal 0, o.illustration_IP_votes.size
     o.illustration_IP_vote = '127.0.0.1'
     o.save
     assert_equal 1, o.illustration_IP_votes.size
     assert_equal ['127.0.0.1'], o.illustration_IP_votes
  end

  test "add illustration IP votes is initialized on create" do
     o = OntologyClass.create!(:definition => 'Foos and bars', :written_by => @ref_stub)
     assert_equal 0, o.illustration_IP_votes.size
  end

  test "logical relatives for parents" do 
    _setup_for_tree_logic_tests
    foo = @a.logical_relatives(:direction => :parents) # defaults to children
     
    assert_equal %w(C E F G H I), foo.keys.sort{|a, b| a.definition <=> b.definition}.collect{|p| p.definition}
    assert_equal true, foo[@e] # assert a redundancy
    assert_equal true, foo[@g]

    assert_equal nil, foo[@a]   # @a is not in our result
    assert_equal nil, foo[@b]   # neither is @b
    assert_equal nil, foo[@d]   # neither is @d
    assert_equal false, foo[@f]
    assert_equal false, foo[@h]
    assert_equal false, foo[@i]
  end

  test "tree relatives with mixed relationships" do 
    _setup_for_tree_like_tests
    @r2 = ObjectRelationship.create!(:interaction => "is_a")
    @o18 = OntologyRelationship.create!(:ontology_class1 => @p8, :ontology_class2 => @p1, :object_relationship => @r2)

    assert_equal 8, @p1.tree_relatives(:relationship_type  => [@r.id, @r2.id]).size
    assert_equal 3, @p2.tree_relatives(:relationship_type  => [@r.id, @r2.id]).size
    assert_equal 2, @p3.tree_relatives(:relationship_type  => [@r.id, @r2.id]).size
    assert_equal 0, @p4.tree_relatives(:relationship_type  => [@r.id, @r2.id]).size
    assert_equal 0, @p5.tree_relatives(:relationship_type  => [@r.id, @r2.id]).size
    assert_equal 0, @p6.tree_relatives(:relationship_type  => [@r.id, @r2.id]).size
    assert_equal 0, @p7.tree_relatives(:relationship_type  => [@r.id, @r2.id]).size
    assert_equal 0, @p8.tree_relatives(:relationship_type  => [@r.id, @r2.id]).size

    assert_equal 3, @p1.tree_relatives(:max_depth => 1,:relationship_type  => [@r.id, @r2.id]).size
  end

  test "tree relatives" do
    _setup_for_tree_like_tests
    assert_equal 7, @p1.tree_relatives.size
    assert_equal 3, @p2.tree_relatives.size
    assert_equal 2, @p3.tree_relatives.size
    assert_equal 0, @p4.tree_relatives.size
    assert_equal 0, @p5.tree_relatives.size
    assert_equal 0, @p6.tree_relatives.size
    assert_equal 0, @p7.tree_relatives.size
    assert_equal 0, @p8.tree_relatives.size

    assert_equal 2, @p1.tree_relatives(:max_depth => 1).size
  end

  test "newick_string" do 
    _setup_for_tree_like_tests
    # slap some labels on the classes created above
    %w/A B C D E F G H/.each do |l|
      label = Label.create!(:name => l)
      Sensu.create!(:label => label, :ontology_class => OntologyClass.find_by_definition(l), :ref => @ref_stub) 
    end
    assert_equal "('A',('B',('D','E','F'),'C',('G','H')));", Ontology::Visualize::Newick.newick_string(:ontology_class => @p1, :color => nil)
  end

  test "js_hash" do
    _setup_for_tree_like_tests
    assert_equal 2, @p1.tree_child.size
    assert_equal 3, @p2.tree_child.size
    assert_equal 2, @p3.tree_child.size
    assert_equal 0, @p8.tree_child.size
    assert_equal "{A:{B:{D,E,F},C:{G,H}}}", @p1.js_hash(:key_is_id => false)
  end

  test "logical relatives for children" do
    _setup_for_tree_logic_tests
    foo = @h.logical_relatives(:direction => :children) # defaults to children
    assert_equal %w(A C F), foo.keys.sort{|a, b| a.definition <=> b.definition}.collect{|p| p.definition}
  end

  ## helpers
  
  def _setup_for_tree_like_tests
    # a
    # b    c
    # def  gh
    @p1 = OntologyClass.create!(:definition => "A", :written_by => @ref_stub)
    @p2 = OntologyClass.create!(:definition => "B", :written_by => @ref_stub)
    @p3 = OntologyClass.create!(:definition => "C", :written_by => @ref_stub)
    @p4 = OntologyClass.create!(:definition => "D", :written_by => @ref_stub)
    @p5 = OntologyClass.create!(:definition => "E", :written_by => @ref_stub)
    @p6 = OntologyClass.create!(:definition => "F", :written_by => @ref_stub)
    @p7 = OntologyClass.create!(:definition => "G", :written_by => @ref_stub)
    @p8 = OntologyClass.create!(:definition => "H", :written_by => @ref_stub)
     
    @r = ObjectRelationship.create!(:interaction => "part_of")
    @r2 = ObjectRelationship.create!(:interaction => "is_a")

    @o12 = OntologyRelationship.create!(:ontology_class1_id => @p2.id, :ontology_class2_id => @p1.id, :object_relationship_id => @r.id)
    @o13 = OntologyRelationship.create!(:ontology_class1_id => @p3.id, :ontology_class2_id => @p1.id, :object_relationship_id => @r.id)
                                                                                 
    @o24 = OntologyRelationship.create!(:ontology_class1_id => @p4.id, :ontology_class2_id => @p2.id, :object_relationship_id => @r.id)
    @o25 = OntologyRelationship.create!(:ontology_class1_id => @p5.id, :ontology_class2_id => @p2.id, :object_relationship_id => @r.id)
    @o26 = OntologyRelationship.create!(:ontology_class1_id => @p6.id, :ontology_class2_id => @p2.id, :object_relationship_id => @r.id)
                                                                                 
    @o37 = OntologyRelationship.create!(:ontology_class1_id => @p7.id, :ontology_class2_id => @p3.id, :object_relationship_id => @r.id)
    @o38 = OntologyRelationship.create!(:ontology_class1_id => @p8.id, :ontology_class2_id => @p3.id, :object_relationship_id => @r.id)
  end

  def _setup_for_tree_logic_tests
    # * -> is a 
    # @ -> part of
    #                    a
    #                 *     @  
    #                b       c
    #              *   @   *   @ 
    #              d   e   f   g
    #                  *   @  
    #                  i   h  
   
    # AND two redundancies : a*e, a*g

    @a = OntologyClass.create!(:definition => "A", :written_by => @ref_stub)
    @b = OntologyClass.create!(:definition => "B", :written_by => @ref_stub)
    @c = OntologyClass.create!(:definition => "C", :written_by => @ref_stub)
    @d = OntologyClass.create!(:definition => "D", :written_by => @ref_stub)
    @e = OntologyClass.create!(:definition => "E", :written_by => @ref_stub)
    @f = OntologyClass.create!(:definition => "F", :written_by => @ref_stub)
    @g = OntologyClass.create!(:definition => "G", :written_by => @ref_stub)
    @h = OntologyClass.create!(:definition => "H", :written_by => @ref_stub)
    @i = OntologyClass.create!(:definition => "I", :written_by => @ref_stub)

    @rp = ObjectRelationship.create!(:interaction => "part_of")
    @ri = ObjectRelationship.create!(:interaction => "is_a")

    @oAB = OntologyRelationship.create!(:ontology_class1 => @a, :ontology_class2 => @b, :object_relationship => @ri)
    @oAC = OntologyRelationship.create!(:ontology_class1 => @a, :ontology_class2 => @c, :object_relationship => @rp)
    @oBD = OntologyRelationship.create!(:ontology_class1 => @b, :ontology_class2 => @d, :object_relationship => @ri)
    @oBE = OntologyRelationship.create!(:ontology_class1 => @b, :ontology_class2 => @e, :object_relationship => @rp)
    @oEI = OntologyRelationship.create!(:ontology_class1 => @e, :ontology_class2 => @i, :object_relationship => @ri)
    @oCF = OntologyRelationship.create!(:ontology_class1 => @c, :ontology_class2 => @f, :object_relationship => @ri)
    @oCG = OntologyRelationship.create!(:ontology_class1 => @c, :ontology_class2 => @g, :object_relationship => @rp)
    @oFH = OntologyRelationship.create!(:ontology_class1 => @f, :ontology_class2 => @h, :object_relationship => @rp)

    # two (logical) redundancies

    @oAE = OntologyRelationship.create!(:ontology_class1 => @a, :ontology_class2 => @e, :object_relationship => @rp)
    @oAG = OntologyRelationship.create!(:ontology_class1 => @a, :ontology_class2 => @g, :object_relationship => @rp)
  end

end
